/*
  4WD Mecanum Wheel Robot Car Base - Simple Maneuver Test
  mecanum-test.ino
  Cycles 4WD Mecanum Wheel Robot Car through simple manuuvers
  Uses ESP32 DevKitC (other ESP32 boards will work)
  Uses 2 TB6612FNG Motor Drivers, also compatible with L298N
  
  DroneBot Workshop 2022
  https://dronebotworkshop.com
*/

// Define Motor Connections
// Right Front Motor
#define MF_PWMA 19
#define MF_AI1 32
#define MF_AI2 23

// Left Front Motor
#define MF_PWMB 26
#define MF_BI1 33
#define MF_BI2 25

// Right Rear Motor
#define MR_PWMA 27
#define MR_AI1 12
#define MR_AI2 14

// Left Rear Motor
#define MR_PWMB 4
#define MR_BI1 13
#define MR_BI2 2

// Define PWM Motor Speed Variables
int rf_PWM = 0;  // Right Front Motor
int lf_PWM = 0;  // Left Front Motor
int rr_PWM = 0;  // Right Rear Motor
int lr_PWM = 0;  // Left Front Motor

// Define Bytes to represent Mecannum Wheel Modes
// Individual bits define TB6612FNG motor driver module input states
// B7 = MF_AI1, B6 = MF_AI2, B5 = MF_BI1, B4 = MF_BI2, B3 = MR_AI1, B2 = MR_AI2, B1 = MR_BI1, B0 = MR_BI2

const byte MEC_STRAIGHT_FORWARD = B10101010;
const byte MEC_STRAIGHT_BACKWARD = B01010101;
const byte MEC_SIDEWAYS_RIGHT = B01101001;
const byte MEC_SIDEWAYS_LEFT = B10010110;
const byte MEC_DIAGONAL_45 = B00101000;
const byte MEC_DIAGONAL_135 = B10000010;
const byte MEC_DIAGONAL_225 = B00010100;
const byte MEC_DIAGONAL_315 = B01000001;
const byte MEC_PIVOT_RIGHT_FORWARD = B00100010;
const byte MEC_PIVOT_RIGHT_BACKWARD = B00010001;
const byte MEC_PIVOT_LEFT_FORWARD = B10001000;
const byte MEC_PIVOT_LEFT_BACKWARD = B01000100;
const byte MEC_ROTATE_CLOCKWISE = B01100110;
const byte MEC_ROTATE_COUNTERCLOCKWISE = B10011001;
const byte MEC_PIVOT_SIDEWAYS_FRONT_RIGHT = B01100000;
const byte MEC_PIVOT_SIDEWAYS_FRONT_LEFT = B10010000;
const byte MEC_PIVOT_SIDEWAYS_REAR_RIGHT = B00001001;
const byte MEC_PIVOT_SIDEWAYS_REAR_LEFT = B00000110;

// Variable for test time delay
int timeDelay = 1000;

// PWM Parameters for motor control
// PWM Frequency = 1KHz
const int mtrPWMFreq = 1000;
// PWM Resolution
const int mtrPWMResolution = 8;
// Define PWM channels for each motor
const int mtrRFpwmchannel = 4;
const int mtrLFpwmchannel = 5;
const int mtrRRpwmchannel = 6;
const int mtrLRpwmchannel = 7;

void moveMotors(int speedRF, int speedLF, int speedRR, int speedLR, byte dircontrol) {

  // Moves all 4 motors
  // Directions specified in direction byte

  // Right Front Motor
  digitalWrite(MF_AI1, bitRead(dircontrol, 7));
  digitalWrite(MF_AI2, bitRead(dircontrol, 6));
  ledcWrite(mtrRFpwmchannel, abs(speedRF));

  // Left Front Motor
  digitalWrite(MF_BI1, bitRead(dircontrol, 5));
  digitalWrite(MF_BI2, bitRead(dircontrol, 4));
  ledcWrite(mtrLFpwmchannel, abs(speedLF));

  // Right Rear Motor
  digitalWrite(MR_AI1, bitRead(dircontrol, 3));
  digitalWrite(MR_AI2, bitRead(dircontrol, 2));
  ledcWrite(mtrRRpwmchannel, abs(speedRR));

  // Left Rear Motor
  digitalWrite(MR_BI1, bitRead(dircontrol, 1));
  digitalWrite(MR_BI2, bitRead(dircontrol, 0));
  ledcWrite(mtrLRpwmchannel, abs(speedLR));
}

void stopMotors() {

  // Stops all motors and motor controllers
  ledcWrite(mtrRFpwmchannel, 0);
  ledcWrite(mtrLFpwmchannel, 0);
  ledcWrite(mtrRRpwmchannel, 0);
  ledcWrite(mtrLRpwmchannel, 0);

  digitalWrite(MF_AI1, 0);
  digitalWrite(MF_AI2, 0);
  digitalWrite(MF_BI1, 0);
  digitalWrite(MF_BI2, 0);
  digitalWrite(MR_AI1, 0);
  digitalWrite(MR_AI2, 0);
  digitalWrite(MR_BI1, 0);
  digitalWrite(MR_BI2, 0);
}

void setup() {

  // Set up Serial Monitor
  Serial.begin(9600);

  // Set all connections as outputs
  pinMode(MF_PWMA, OUTPUT);
  pinMode(MF_AI1, OUTPUT);
  pinMode(MF_AI2, OUTPUT);
  pinMode(MF_PWMB, OUTPUT);
  pinMode(MF_BI1, OUTPUT);
  pinMode(MF_BI2, OUTPUT);
  pinMode(MR_PWMA, OUTPUT);
  pinMode(MR_AI1, OUTPUT);
  pinMode(MR_AI2, OUTPUT);
  pinMode(MR_PWMB, OUTPUT);
  pinMode(MR_BI1, OUTPUT);
  pinMode(MR_BI2, OUTPUT);

  //Set up PWM channels with frequency and resolution
  ledcSetup(mtrRFpwmchannel, mtrPWMFreq, mtrPWMResolution);
  ledcSetup(mtrLFpwmchannel, mtrPWMFreq, mtrPWMResolution);
  ledcSetup(mtrRRpwmchannel, mtrPWMFreq, mtrPWMResolution);
  ledcSetup(mtrLRpwmchannel, mtrPWMFreq, mtrPWMResolution);

  // Attach channels to PWM output pins
  ledcAttachPin(MF_PWMA, mtrRFpwmchannel);
  ledcAttachPin(MF_PWMB, mtrLFpwmchannel);
  ledcAttachPin(MR_PWMA, mtrRRpwmchannel);
  ledcAttachPin(MR_PWMB, mtrLRpwmchannel);

  // Test speed for all motors (change as desired)
  rf_PWM = 200;  // Right Front Motor
  lf_PWM = 200;  // Left Front Motor
  rr_PWM = 200;  // Right Rear Motor
  lr_PWM = 200;  // Left Front Motor
}

void loop() {

  // Cycle through some Mecanum Wheel modes

  delay(3000);  

  // Straight Forward
  Serial.println("Straight Forward");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_STRAIGHT_FORWARD);
  delay(timeDelay);
  stopMotors();
  delay(500);

  // Straight Backward
  Serial.println("Straight Backward");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_STRAIGHT_BACKWARD);
  delay(timeDelay);
  stopMotors();
  delay(500);

  // Sideways Right
  Serial.println("Sideways Right");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_SIDEWAYS_RIGHT);
  delay(timeDelay);
  stopMotors();
  delay(500);

  // Sideways Left
  Serial.println("Sideways Left");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_SIDEWAYS_LEFT);
  delay(timeDelay);
  stopMotors();
  delay(500);

  // Diagonal 45 Degrees
  Serial.println("Diagonal 45 Degrees");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_DIAGONAL_45);
  delay(timeDelay);
  stopMotors();
  delay(500);

  // Diagonal 225 Degrees
  Serial.println("Diagonal 225 Degrees");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_DIAGONAL_225);
  delay(timeDelay);
  moveMotors(0, 0, 0, 0, MEC_DIAGONAL_225);
  delay(500);

  // Diagonal 135 Degrees
  Serial.println("Diagonal 135 Degrees");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_DIAGONAL_135);
  delay(timeDelay);
  moveMotors(0, 0, 0, 0, MEC_DIAGONAL_135);
  delay(500);

  // Diagonal 315 Degrees
  Serial.println("Diagonal 315 Degrees");
  moveMotors(rf_PWM, lf_PWM, rr_PWM, lr_PWM, MEC_DIAGONAL_315);
  delay(timeDelay);
  moveMotors(0, 0, 0, 0, MEC_STRAIGHT_FORWARD);
  delay(500);
}